from __future__ import annotations

import typing

import absorb
from . import cli_helpers

if typing.TYPE_CHECKING:
    import argparse


def get_subcommands() -> (
    list[tuple[str, str, list[tuple[list[str], dict[str, typing.Any]]]]]
):
    return [
        (
            'ls',
            'list tracked datasets',
            [
                (
                    ['--verbose', '-v'],
                    {
                        'action': 'store_true',
                        'help': 'show verbose details',
                    },
                ),
            ],
        ),
        (
            'help',
            'show info about a specific dataset or data source',
            [
                (
                    ['dataset'],
                    {
                        'help': 'dataset or data source',
                    },
                )
            ],
        ),
        (
            'collect',
            'collect datasets',
            [
                (
                    ['dataset'],
                    {
                        'nargs': '*',
                        'help': 'dataset to track, format as "<source>.<dataset>"',
                    },
                ),
                (
                    ['--parameters'],
                    {'nargs': '*', 'help': 'dataset parameters'},
                ),
                (
                    ['--dry'],
                    {
                        'action': 'store_true',
                        'help': 'perform dry run (avoids collecting data)',
                    },
                ),
                (
                    ['--overwrite'],
                    {
                        'action': 'store_true',
                        'help': 'overwrite existing files',
                    },
                ),
                (
                    ['--range'],
                    {
                        'help': 'range of data to collect',
                    },
                ),
                (
                    ['-v', '--verbose'],
                    {
                        'help': 'display extra information',
                        'nargs': '?',
                        'const': 1,
                        'default': 1,
                        'type': int,
                    },
                ),
            ],
        ),
        (
            'add',
            'start tracking datasets',
            [
                (
                    ['dataset'],
                    {
                        'nargs': '*',
                        'help': 'dataset to track, format as "<source>.<dataset>"',
                    },
                ),
                (
                    ['--parameters'],
                    {'nargs': '*', 'help': 'dataset parameters'},
                ),
                (
                    ['--all'],
                    {
                        'help': 'add all available datasets',
                        'action': 'store_true',
                    },
                ),
                (
                    ['--path'],
                    {'help': 'directory location to store the dataset'},
                ),
            ],
        ),
        (
            'remove',
            'remove tracking datasets',
            [
                (
                    ['dataset'],
                    {
                        'nargs': '*',
                        'help': 'dataset to track, format as "<source>.<dataset>"',
                    },
                ),
                (
                    ['--parameters'],
                    {'nargs': '*', 'help': 'dataset parameters'},
                ),
                (
                    ['--all'],
                    {
                        'help': 'add all available datasets',
                        'action': 'store_true',
                    },
                ),
            ],
        ),
        (
            'path',
            'print absorb root path or dataset path',
            [
                (
                    ['dataset'],
                    {
                        'nargs': '?',
                        'help': 'dataset to track, format as "<source>.<dataset>"',
                    },
                ),
                (
                    ['--parameters'],
                    {'nargs': '*', 'help': 'dataset parameters'},
                ),
                (
                    ['--glob'],
                    {'action': 'store_true'},
                ),
            ],
        ),
        (
            'new',
            'create new dataset',
            [
                (
                    ['dataset'],
                    {
                        'nargs': '?',
                        'help': 'dataset to track, format as "<source>.<dataset>"',
                    },
                ),
                (
                    ['--path'],
                    {
                        'help': 'path where to store new table definition',
                    },
                ),
                (
                    ['--native'],
                    {
                        'action': 'store_true',
                        'help': 'create definition directly in absorb repo',
                    },
                ),
            ],
        ),
    ]


def parse_args() -> argparse.Namespace:
    import argparse
    import importlib

    parser = argparse.ArgumentParser(
        formatter_class=cli_helpers.HelpFormatter, allow_abbrev=False
    )
    subparsers = parser.add_subparsers(dest='command')

    parsers = {}
    for name, description, arg_args in get_subcommands():
        f_module = importlib.import_module(
            'absorb.cli.cli_commands.command_' + name
        )
        f = getattr(f_module, name + '_command')
        subparser = subparsers.add_parser(name, help=description)
        subparser.set_defaults(f_command=f)
        for sub_args, sub_kwargs in arg_args:
            subparser.add_argument(*sub_args, **sub_kwargs)
        subparser.add_argument(
            '--debug',
            '--pdb',
            help='enter debugger upon error',
            action='store_true',
        )
        subparser.add_argument(
            '-i',
            '--interactive',
            help='open data in interactive python session',
            action='store_true',
        )
        parsers[name] = subparser

    # parse args
    args = parser.parse_args()

    # display help if no command specified
    if args.command is None:
        import sys

        parser.print_help()
        sys.exit(0)

    return args


def _parse_datasets(args: argparse.Namespace) -> list[absorb.TrackedTable]:
    # parse datasets
    sources = []
    tables = []
    classes = []
    if isinstance(args.dataset, list):
        datasets = args.dataset
    elif isinstance(args.dataset, str):
        datasets = [args.dataset]
    else:
        raise Exception()
    for dataset in datasets:
        # get source and table
        if '.' in dataset:
            source, table = dataset.split('.')
            sources.append(source)
            tables.append(table)
        else:
            for source_dataset in absorb.ops.get_source_tables(dataset):
                sources.append(dataset)
                tables.append(source_dataset.__name__)

        # get table class
        table_class = absorb.ops.get_table_class(
            source=sources[-1], table_name=tables[-1]
        )
        classes.append(table_class)

    # parse parameter types
    parameter_types = {}
    for table_class in classes:
        for parameter, parameter_type in table_class.parameter_types.items():
            if parameter not in parameter_types:
                parameter_types[parameter] = parameter_type
            else:
                if parameter_type != parameter_types[parameter]:
                    raise Exception(
                        'inconsistent parameter types across datasets'
                    )

    # parse parameters
    parameters: dict[str, typing.Any] = {}
    value: typing.Any
    if args.parameters is not None:
        for parameter in args.parameters:
            key, value = parameter.split('=')

            # set parameter type
            if key not in parameter_types:
                raise Exception(
                    'unknown parameter: '
                    + str(key)
                    + ' not in '
                    + str(list(parameters.keys()))
                )
            parameter_type = parameter_types[key]
            if parameter_type == str:
                pass
            elif parameter_type == int:
                value = int(value)
            elif parameter_type == list[str]:
                value = value.split(',')
            elif parameter_type == list[int]:
                value = [int(subvalue) for subvalue in value.split(',')]
            else:
                raise Exception(
                    'invalid parameter type: ' + str(parameter_type)
                )

            parameters[key] = value

    # create TrackedTable dicts
    parsed = []
    for source, table in zip(sources, tables):
        camel_table = absorb.ops.names._snake_to_camel(table)
        tracked_table: absorb.TrackedTable = {
            'source_name': source,
            'table_name': table,
            'table_class': 'absorb.datasets.' + source + '.' + camel_table,
            'parameters': parameters,
        }
        parsed.append(tracked_table)

    return parsed


def _parse_ranges(
    raw_ranges: list[str] | None, range_format: absorb.RangeFormat
) -> list[typing.Any] | None:
    """
    examples:
    --range 2025-01-01:2025-03-01
    --range 2025-01-01:
    --range :2025-01-01
    """
    if raw_ranges is None:
        return None
    if range_format == 'date' or range_format == 'date_range':
        raise NotImplementedError('manual ranges for ' + str(range_format))
    else:
        raise NotImplementedError('manual ranges for ' + str(range_format))

    # 'date',
    # 'date_range',
    # 'named_range',
    # 'block_range',
    # 'id_range',
    # 'count',
    # None,



from __future__ import annotations

from . import cli_parsing
from . import cli_helpers


def run_cli() -> None:
    args = cli_parsing.parse_args()
    try:
        data = args.f_command(args)
        if args.interactive:
            cli_helpers.open_interactive_session(variables=data)
    except BaseException as e:
        if args.debug:
            cli_helpers._enter_debugger()
        else:
            raise e



from __future__ import annotations

import typing

import absorb
from .. import cli_outputs

if typing.TYPE_CHECKING:
    from argparse import Namespace
    from typing import Any


def ls_command(args: Namespace) -> dict[str, Any]:
    import toolstr

    tracked_datasets = absorb.ops.get_tracked_tables()

    # available datasets
    cli_outputs._print_title('Available datasets')
    for source in sorted(absorb.ops.get_sources()):
        source_tables = absorb.ops.get_source_tables(source)
        if len(source_tables) > 0:
            cli_outputs._print_source_datasets_bullet(source, source_tables)

    # tracked datasets
    print()
    if args.verbose:
        cli_outputs._print_title('Tracked datasets')
        if len(tracked_datasets) == 0:
            print('[none]')
        else:
            rows = []
            for dataset in tracked_datasets:
                instance = absorb.Table.instantiate(dataset)
                available_range = instance.get_available_range()
                available_range = absorb.ops.format_range(available_range)
                collected_range = instance.get_collected_range()
                if collected_range is not None:
                    collected_range = absorb.ops.format_range(collected_range)
                else:
                    collected_range = '-'
                row = [
                    dataset['source_name'],
                    dataset['table_name'],
                    available_range + '\n' + collected_range,
                ]
                rows.append(row)
            columns = ['source', 'table', 'available range\ncollected range']
            toolstr.print_multiline_table(rows, labels=columns)
    else:
        cli_outputs._print_title('Tracked datasets')
        if len(tracked_datasets) == 0:
            print('[none]')
        else:
            for dataset in tracked_datasets:
                cli_outputs._print_dataset_bullet(dataset)

    return {}



from __future__ import annotations

import absorb


def _print_title(title: str) -> None:
    import rich

    rich.print('[bold green]' + title + '[/bold green]')


def _dataset_to_str(dataset: absorb.TrackedTable) -> str:
    return dataset['source_name'] + '.' + dataset['table_name']


def _print_dataset_bullet(dataset: absorb.TrackedTable) -> None:
    import toolstr

    toolstr.print_bullet(
        '[white bold]' + _dataset_to_str(dataset) + '[/white bold]',
        **absorb.ops.formatting.bullet_styles,
    )


def _print_source_datasets_bullet(
    source: str, datasets: list[type[absorb.Table]]
) -> None:
    import toolstr

    names = [cls.class_name(allow_generic=True) for cls in datasets]
    toolstr.print_bullet(
        key=source,
        value='[green],[/green] '.join(names),
        **absorb.ops.formatting.bullet_styles,
    )



from __future__ import annotations

import typing
from . import table_class

if typing.TYPE_CHECKING:
    from typing_extensions import NotRequired


RangeFormat = typing.Literal[
    'date',
    'hour',
    'date_range',
    'per_hour',
    'named_range',
    'block_range',
    'id_range',
    'count',
    None,
]


class Context(typing.TypedDict):
    parameters: dict[str, typing.Any]
    data_range: typing.Any
    overwrite: bool


class TrackedTable(typing.TypedDict):
    source_name: str
    table_name: str
    table_class: str
    parameters: dict[str, typing.Any]


TableReference = typing.Union[str, TrackedTable, table_class.Table]


class Config(typing.TypedDict):
    version: str
    tracked_tables: list[TrackedTable]


class NameTemplate(typing.TypedDict):
    default: NotRequired[str]
    custom: NotRequired[str | dict[str | tuple[str], str]]



from __future__ import annotations

import typing

import absorb


_cache = {'root_dir_warning_shown': False}


def get_absorb_root(*, warn: bool = False) -> str:
    import os

    path = os.environ.get('TRUCK_ROOT')
    if path is None or path == '':
        if warn and not _cache['root_dir_warning_shown']:
            import rich

            rich.print(
                '[#777777]using default value for TRUCK_ROOT: ~/absorb[/#777777]\n'
            )
            _cache['root_dir_warning_shown'] = True
        path = '~/absorb'
    path = os.path.expanduser(path)
    return path


def get_config_path(*, warn: bool = False) -> str:
    import os

    return os.path.join(
        absorb.ops.get_absorb_root(warn=warn), 'absorb_config.json'
    )


def get_source_dir(source: str, *, warn: bool = False) -> str:
    import os

    return os.path.join(get_absorb_root(warn=warn), 'datasets', source)


def get_table_dir(
    dataset: absorb.TrackedTable | str | None = None,
    *,
    source: str | None = None,
    table: str | None = None,
    warn: bool = False,
) -> str:
    import os

    if isinstance(dataset, str):
        source, table = dataset.split('.')
    elif isinstance(dataset, dict):
        source = dataset['source_name']
        table = dataset['table_name']
    elif source is not None and table is not None:
        pass
    else:
        raise Exception('invalid format')

    source_dir = get_source_dir(source, warn=warn)
    return os.path.join(source_dir, 'tables', table)


def get_table_metadata_path(
    dataset: absorb.TrackedTable | str, *, warn: bool = False
) -> str:
    import os

    return os.path.join(
        absorb.ops.get_table_dir(dataset, warn=warn), 'table_metadata.json'
    )


def get_table_filepath(
    data_range: typing.Any,
    range_format: absorb.RangeFormat,
    filename_template: str,
    table: str,
    *,
    source: str | None,
    parameters: dict[str, typing.Any],
    glob: bool = False,
    warn: bool = True,
) -> str:
    import os

    dir_path = get_table_dir(source=source, table=table, warn=warn)
    filename = get_table_filename(
        data_range=data_range,
        range_format=range_format,
        filename_template=filename_template,
        table=table,
        source=source,
        parameters=parameters,
        glob=glob,
    )
    return os.path.join(dir_path, filename)


def get_table_filename(
    data_range: typing.Any,
    range_format: absorb.RangeFormat,
    filename_template: str,
    table: str,
    *,
    source: str | None,
    parameters: dict[str, typing.Any],
    glob: bool = False,
) -> str:
    format_params = parameters.copy()
    if source is not None:
        format_params['source'] = source
    format_params['table'] = table
    if '{data_range}' in filename_template:
        if glob:
            format_params['data_range'] = '*'
        else:
            format_params['data_range'] = _format_data_range(
                data_range, range_format
            )
    return filename_template.format(**format_params)


def get_table_filepaths(
    data_ranges: typing.Any,
    range_format: absorb.RangeFormat,
    filename_template: str,
    table: str,
    *,
    source: str | None,
    parameters: dict[str, typing.Any],
    warn: bool = True,
) -> list[str]:
    import os

    dir_path = get_table_dir(source=source, table=table, warn=warn)
    paths = []
    for data_range in data_ranges:
        filename = get_table_filename(
            data_range=data_range,
            range_format=range_format,
            filename_template=filename_template,
            table=table,
            source=source,
            parameters=parameters,
        )
        path = os.path.join(dir_path, filename)
        paths.append(path)
    return paths


def _format_data_range(
    data_range: typing.Any, range_format: absorb.RangeFormat
) -> str:
    if range_format == 'date':
        return data_range.strftime('%Y-%m-%d')  # type: ignore
    elif range_format == 'date_range':
        return (
            _format_value(data_range[0], 'date')
            + '_to_'
            + _format_value(data_range[1], 'date')
        )
    elif range_format == 'named_range':
        if len(data_range) == 2:
            values = list(data_range.values())
            values = sorted(values)
            values_str = [_format_value(value) for value in values]
            return values_str[0] + '_to_' + values_str[1]
        else:
            raise NotImplementedError('range with >2 keys')
    elif range_format == 'block_range':
        return (
            _format_value(data_range[0], 'int')
            + '_to_'
            + _format_value(data_range[1], 'int')
        )
    elif range_format == 'id_range':
        return (
            _format_value(data_range[0]) + '_to_' + _format_value(data_range[1])
        )
    elif range_format is None:
        return str(data_range)
    else:
        raise Exception('invalid range_format: ' + str(range_format))


def _format_value(
    value: typing.Any, format: str | None = None, width: int = 10
) -> str:
    if format is None:
        import datetime

        if isinstance(value, datetime.datetime):
            format = 'date'
        elif isinstance(value, int):
            format = 'int'
        else:
            raise Exception('invalid format')

    if format == 'date':
        return value.strftime('%Y-%m-%d')  # type: ignore
    elif format == 'int':
        return ('%0' + str(width) + 'd') % value  # type: ignore
    else:
        raise Exception('invalid format')


def parse_file_path(
    path: str,
    filename_template: str,
    *,
    range_format: absorb.RangeFormat | None = None,
) -> dict[str, typing.Any]:
    import os

    keys = os.path.splitext(filename_template)[0].split('__')
    values = os.path.splitext(os.path.basename(path))[0].split('__')
    items = {k[1:-1]: v for k, v in zip(keys, values)}
    if range_format is not None and 'data_range' in items:
        items['data_range'] = parse_data_range(
            items['data_range'], range_format
        )
    return items


def parse_data_range(
    as_str: str, range_format: absorb.RangeFormat
) -> typing.Any:
    if range_format == 'date':
        import datetime

        return datetime.datetime.strptime(as_str, '%Y-%m-%d')
    else:
        raise NotImplementedError()



from __future__ import annotations

import typing
import absorb

from . import table_coverage

if typing.TYPE_CHECKING:
    T = typing.TypeVar('T')

    import polars as pl


class TableCollect(table_coverage.TableCoverage):
    def collect_chunk(self, data_range: typing.Any) -> pl.DataFrame | None:
        raise NotImplementedError()

    def collect(
        self,
        data_range: typing.Any | None = None,
        *,
        overwrite: bool = False,
        verbose: int = 1,
        dry: bool = False,
    ) -> None:
        # get collection plan
        chunk_ranges, paths = self._get_chunks_to_collect(data_range, overwrite)
        if not overwrite:
            chunk_ranges, paths = self._filter_existing_chunks(
                chunk_ranges=chunk_ranges, paths=paths
            )

        # summarize collection
        if verbose >= 1:
            self._summarize_collection_plan(
                chunk_ranges=chunk_ranges,
                paths=paths,
                overwrite=overwrite,
                verbose=verbose,
            )

        # return early if dry
        if dry:
            return None

        # collect each chunk
        for chunk_range, path in zip(chunk_ranges, paths):
            if verbose >= 1:
                self._summarize_chunk(chunk_range, path)
            df = self.collect_chunk(data_range=chunk_range)
            if df is not None:
                absorb.ops.write_file(df=df, path=path)
            if verbose >= 1:
                self._summarize_collected_chunk(df, chunk_range, path)

    def _get_chunks_to_collect(
        self, data_range: typing.Any | None = None, overwrite: bool = False
    ) -> tuple[list[typing.Any], list[str]]:
        if self.write_range == 'overwrite_all':
            chunk_ranges = [None]
            paths = [self.get_file_path(None)]
        else:
            if data_range is None:
                data_ranges = self._get_missing_data_ranges(overwrite)
            else:
                data_ranges = [data_range]
            chunk_ranges = absorb.ops.ranges.partition_into_chunks(
                data_ranges, range_format=self.range_format
            )
            paths = [
                self.get_file_path(data_range=chunk_range)
                for chunk_range in chunk_ranges
            ]
        return chunk_ranges, paths

    def _get_missing_data_ranges(self, overwrite: bool) -> list[typing.Any]:
        if overwrite:
            return [self.get_available_range()]
        else:
            collected_range = self.get_collected_range()
            available_range = self.get_available_range()
            if collected_range is None:
                return [available_range]
            else:
                return absorb.ops.ranges.get_range_diff(
                    subtract_this=collected_range,
                    from_this=available_range,
                    range_format=self.range_format,
                )

    def _filter_existing_chunks(
        self, chunk_ranges: list[typing.Any], paths: list[str]
    ) -> tuple[list[typing.Any], list[str]]:
        import os

        new_chunk_ranges: list[typing.Any] = []
        new_paths = []
        for chunk_range, path in zip(chunk_ranges, paths):
            if not os.path.isfile(path):
                new_chunk_ranges.append(chunk_range)
                new_paths.append(path)
        return new_chunk_ranges, new_paths

    def _summarize_collection_plan(
        self,
        chunk_ranges: list[typing.Any],
        paths: list[str],
        overwrite: bool,
        verbose: int,
    ) -> None:
        import datetime
        import rich

        rich.print(
            '[bold][green]collecting dataset:[/green] [white]'
            + self.source
            + '.'
            + self.name()
            + '[/white][/bold]'
        )
        absorb.ops.print_bullet('n_chunks', str(len(chunk_ranges)))
        if self.write_range == 'overwrite_all':
            absorb.ops.print_bullet('chunk', '\[entire dataset]')  # noqa
        elif len(chunk_ranges) == 1:
            absorb.ops.print_bullet('single chunk', chunk_ranges[0])
        elif len(chunk_ranges) > 1:
            absorb.ops.print_bullet('min_chunk', chunk_ranges[0])
            absorb.ops.print_bullet('max_chunk', chunk_ranges[-1])
        absorb.ops.print_bullet('overwrite', str(overwrite))
        absorb.ops.print_bullet('output dir', self.get_dir_path())
        absorb.ops.print_bullet(
            'collection start time', str(datetime.datetime.now())
        )
        if len(chunk_ranges) == 0:
            print('[already collected]')
        print()

        if verbose > 1:
            absorb.ops.print_bullet(key='chunks', value='')
            for c, chunk_range in enumerate(chunk_ranges):
                absorb.ops.print_bullet(
                    key=None,
                    value=absorb.ops.format_range(chunk_range),
                    number=c + 1,
                    indent=4,
                )

    def _summarize_chunk(self, data_range: typing.Any, path: str) -> None:
        import os

        print('collecting', os.path.basename(path))

    def _summarize_collected_chunk(
        self, df: pl.DataFrame | None, chunk_range: typing.Any, path: str
    ) -> None:
        import os

        if df is None:
            print('could not collect data for', os.path.basename(path))



from __future__ import annotations

import typing
import absorb


def _camel_to_snake(name: str) -> str:
    result = []
    for i, char in enumerate(name):
        if char.isupper():
            if i != 0:
                result.append('_')
            result.append(char.lower())
        else:
            result.append(char)
    return ''.join(result)


def _snake_to_camel(name: str) -> str:
    result = []
    capitalize_next = False

    for i, char in enumerate(name):
        if char == '_':
            capitalize_next = True
        elif capitalize_next:
            result.append(char.upper())
            capitalize_next = False
        else:
            result.append(char)

    return result[0].upper() + ''.join(result[1:])


def get_table_name(
    *,
    base_name: str,
    name_template: absorb.types.NameTemplate,
    parameter_types: dict[str, typing.Any],
    parameters: dict[str, typing.Any],
    default_parameters: dict[str, typing.Any],
    static_parameters: list[str],
    allow_generic: bool = False,
) -> str:
    """
    keys
    - base_name: the name of the table datatype
    - {parameter_name}: the name of any of the dataset parameters
    """
    if name_template.get('default') is None:
        pass

    if len(parameter_types) == 0:
        if len(name_template) == 0:
            return base_name
        elif name_template.get('default') is not None:
            template = name_template['default']
        elif name_template.get('custom') is not None:
            if isinstance(name_template['custom'], str):
                template = name_template['custom']
            else:
                raise Exception()
        else:
            raise Exception()
    else:
        parameters_are_default = all(
            parameters.get(key) == default_parameters.get(key)
            for key in parameter_types
        )
        if name_template.get('default') is not None and parameters_are_default:
            template = name_template['default']
        elif name_template.get('custom') is not None:
            custom_template = name_template['custom']
            if isinstance(custom_template, str):
                template = custom_template
            elif isinstance(custom_template, dict):
                for key, value in custom_template.items():
                    if isinstance(key, str):
                        variables: str | tuple[str] = (key,)
                    elif isinstance(key, tuple):
                        variables = key
                    else:
                        raise Exception()
                    if all(variable is not None for variable in variables):
                        template = value
                        break
                else:
                    raise Exception()
            else:
                raise Exception()
        elif len(name_template) == 0 and parameters_are_default:
            return base_name
        else:
            raise Exception('need parameters specified')

    if allow_generic:
        template_items = dict(base_name=base_name, **parameters)
        result = template
        for key, value in template_items.items():
            result = result.replace('{' + key + '}', str(value))
        return result
    else:
        return template.format(base_name=base_name, **parameters)



# class DataRange(typing.TypedDict):
#     """specify data range using one of:
#     1. time bounds
#     2. non-temporal bounds
#     3. list of chunks
#     """

#     start_time: datetime.datetime | None
#     end_time: datetime.datetime | None
#     start_bound: typing.Any | None
#     end_bound: typing.Any | None
#     chunks: list[typing.Any] | None


# def get_path(table: absorb.TableReference, date: datetime.datetime, context: absorb.Context) -> str:
#     path_template = '/Users/stormslivkoff/data/kalshi/raw_archive/market_data_{year}-{month:02}-{day:02}.json'
#     return path_template.format(
#         year=date.year, month=date.month, day=date.day
#     )


# def collect(
#     table: absorb.TableReference,
#     data_range: typing.Any,
#     overwrite: bool = False,
# ) -> None:
#     import tooltime

#     cls = absorb.resolve_table_class(table)

#     if data_range is None:
#         collected = cls.get_collected_range()
#         available = cls.get_available_range()
#         missing = None

#     if cls.cadence == 'daily':
#         dates = tooltime.get_intervals(
#             start='2021-06-28',
#             end=tooltime.now(),
#             interval='1d',
#         )['start']
#         for date in dates:
#             date_context = context.copy()
#             date_context['data_range'] = date
#             df = cls.collect(data_range=data_range)
#             path = cls.get_path(date_context)
#             df.write_parquet(path)
#     else:
#         raise Exception()


# def download(
#     start_time: tooltime.Timestamp,
#     end_time: tooltime.Timestamp,
#     root_dir: str | None = None,
# ) -> None:
#     import os
#     import tooltime

#     # determine timestamps
#     timestamps = tooltime.get_intervals(
#         start=start_time, end=end_time, interval='1d'
#     )['start']

#     # summmarize
#     if root_dir is None:
#         root_dir = paths.get_absorb_root()
#     print('downloading', len(timestamps), 'files to:', root_dir)

#     # download files
#     for timestamp in timestamps:
#         url = get_url(timestamp=timestamp)
#         output_path = paths.get_path(timestamp=timestamp, root_dir=root_dir)
#         if os.path.exists(output_path):
#             print('already downloaded', timestamp)
#         else:
#             print('downloading', timestamp)
#             _download_file(url=url, output_path=output_path)


# def get_url(timestamp: tooltime.Timestamp) -> str:
#     return paths.get_path(
#         timestamp=timestamp,
#         root_dir='https://mempool-dumpster.flashbots.net/ethereum/mainnet',
#         flat=False,
#     )


# def _download_file(url: str, output_path: str) -> None:
#     """generic downloader function"""
#     import os
#     import requests
#     import shutil

#     os.makedirs(os.path.dirname(output_path), exist_ok=True)
#     response = requests.get(url, stream=True)
#     response.raise_for_status()
#     tmp_path = output_path + '_tmp'
#     with open(tmp_path, 'wb') as file:
#         for chunk in response.iter_content(chunk_size=8192):
#             if chunk:
#                 file.write(chunk)
#     shutil.move(tmp_path, output_path)



from __future__ import annotations

import typing

import absorb

if typing.TYPE_CHECKING:
    import datetime


# def get_min_local_timestamp(
#     dataset: absorb.TableReference, context: absorb.Context | None = None
# ) -> datetime.datetime:
#     return absorb.resolve_table_class(dataset).get_min_local_timestamp(context)


# def get_max_local_timestamp(
#     dataset: absorb.TableReference, context: absorb.Context | None = None
# ) -> datetime.datetime:
#     return absorb.resolve_table_class(dataset).get_max_local_timestamp(context)


# def get_min_available_timestamp(
#     dataset: absorb.TableReference, context: absorb.Context | None = None
# ) -> datetime.datetime:
#     return absorb.resolve_table_class(dataset).get_min_available_timestamp(
#         context
#     )


# def get_max_available_timestamp(
#     dataset: absorb.TableReference, context: absorb.Context | None = None
# ) -> datetime.datetime:
#     return absorb.resolve_table_class(dataset).get_max_available_timestamp(
#         context
#     )



from __future__ import annotations

import typing

import absorb

if typing.TYPE_CHECKING:
    import polars as pl


class Query(absorb.Table):
    source = 'snowflake'
    write_range = 'overwrite_all'
    range_format = 'per_hour'
    parameters = {'name': str, 'sql': str}

    def get_schema(self) -> dict[str, type[pl.DataType] | pl.DataType]:
        raise NotImplementedError()

    def collect_chunk(self, data_range: typing.Any) -> pl.DataFrame:
        import garlic

        sql = self.parameters['sql']
        return garlic.query(sql)

    def get_available_range(self) -> typing.Any:
        raise NotImplementedError()



from __future__ import annotations

import typing
import absorb

if typing.TYPE_CHECKING:
    import datetime

    _T = typing.TypeVar('_T', int, datetime.datetime)


def get_range_diff(
    subtract_this: typing.Any,
    from_this: typing.Any,
    range_format: absorb.RangeFormat,
) -> list[typing.Any]:
    """
    subtraction behaves differently depending on range format
    - mainly, range_format is discrete-closed or continuous-semiopen or other
    - some of these cases will have equivalent outcomes
        - handling them separately keeps maximum clarity + robustness

                                           fs         fe
    original interval                      |----------|
    16 cases of subtraction    1.  |----|
                               2.  |-------|
                               3.  |------------|
                               4.  |------------------|
                               5.  |------------------------|
                               6.          |
                               7.          |------|
                               8.          |----------|
                               9.          |---------------|
                               10.             |
                               11.             |----|
                               12.             |------|
                               13.             |-----------|
                               14.                    |
                               15.                    |-----|
                               16.                        |----|
                                                          ss   se

    if fs == fe
                                            |
                                1.    |--|
                                2.    |-----|
                                3.    |--------|
                                4.          |
                                5.          |--|
                                6.             |--|
    """
    if range_format == 'date':
        import datetime

        discrete_step = datetime.timedelta(days=1)

        return _get_discrete_closed_range_diff(
            subtract_this=subtract_this,
            from_this=from_this,
            discrete_step=discrete_step,
        )
    elif range_format == 'date_range':
        return _get_continuous_closed_open_range_diff(
            subtract_this=subtract_this,
            from_this=from_this,
        )
    elif range_format == 'named_range':
        assert isinstance(subtract_this, list) and isinstance(from_this, list)
        return [item for item in from_this if item not in set(subtract_this)]
    elif range_format == 'block_range':
        return _get_discrete_closed_range_diff(
            subtract_this=subtract_this, from_this=from_this, discrete_step=1
        )
    elif range_format == 'id_range':
        raise NotImplementedError()
    elif range_format == 'count':
        raise NotImplementedError()
    else:
        raise NotImplementedError()


def _get_discrete_closed_range_diff(
    subtract_this: tuple[_T, _T],
    from_this: tuple[_T, _T],
    discrete_step: typing.Any,
) -> list[tuple[_T, _T]]:
    s_start, s_end = subtract_this
    f_start, f_end = from_this

    # validity checks
    if s_start > s_end:
        raise Exception('invalid interval, start must be <= end')
    if f_start > f_end:
        raise Exception('invalid interval, start must be <= end')

    # 6 possible cases when f_start == f_end
    if f_start == f_end:
        if s_start < f_start and s_end < f_start:
            # case 1
            return [(f_start, f_end)]
        elif s_start < f_start and s_end == f_start:
            # case 2
            return []
        elif s_start < f_start and s_end > f_start:
            # case 3
            return []
        elif s_start == f_start and s_end == f_start:
            # case 4
            return []
        elif s_start == f_start and s_end > f_start:
            # case 5
            return []
        elif s_start > f_start and s_end > f_start:
            # case 6
            return [(f_start, f_end)]
        else:
            raise Exception()

    # 16 possible cases when f_start < f_end
    if s_start < f_start and s_end < f_start:
        # case 1
        return [(f_start, f_end)]
    elif s_start < f_start and s_end == f_start:
        # case 2
        return [(s_end + discrete_step, f_end)]
    elif s_start < f_start and s_end < f_end:
        # case 3
        return [(s_end + discrete_step, f_end)]
    elif s_start < f_start and s_end == f_end:
        # case 4
        return []
    elif s_start < f_start and s_end > f_end:
        # case 5
        return []
    elif s_start == f_start and s_end == f_start:
        # case 6
        return [(s_end + discrete_step, f_end)]
    elif s_start == f_start and s_end < f_end:
        # case 7
        return [(s_end + discrete_step, f_end)]
    elif s_start == f_start and s_end == f_end:
        # case 8
        return []
    elif s_start == f_start and s_end > f_end:
        # case 9
        return []
    elif s_start < f_end and s_end == s_start:
        # case 10
        return [
            (f_start, s_start - discrete_step),
            (s_end + discrete_step, f_end),
        ]
    elif s_start < f_end and s_end < f_end:
        # case 11
        return [
            (f_start, s_start - discrete_step),
            (s_end + discrete_step, f_end),
        ]
    elif s_start < f_end and s_end == f_end:
        # case 12
        return [(f_start, s_start - discrete_step)]
    elif s_start < f_end and s_end > f_end:
        # case 13
        return [(f_start, s_start - discrete_step)]
    elif s_start == f_end and s_end == f_end:
        # case 14
        return [(f_start, s_start - discrete_step)]
    elif s_start == f_end and s_end > f_end:
        # case 15
        return [(f_start, s_start - discrete_step)]
    elif s_start > f_end and s_end > f_start:
        # case 16
        return [(f_start, f_end)]
    else:
        raise Exception()


def _get_continuous_closed_open_range_diff(
    subtract_this: tuple[_T, _T], from_this: tuple[_T, _T]
) -> list[tuple[_T, _T]]:
    s_start, s_end = subtract_this
    f_start, f_end = from_this

    # validity checks
    if s_start >= s_end:
        raise Exception('invalid interval, start must be < end')
    if f_start >= f_end:
        raise Exception('invalid interval, start must be < end')

    # 16 possible cases
    if s_start < f_start and s_end < f_start:
        # case 1
        return [(f_start, f_end)]
    elif s_start < f_start and s_end == f_start:
        # case 2
        return [(f_start, f_end)]
    elif s_start < f_start and s_end < f_end:
        # case 3
        return [(s_end, f_end)]
    elif s_start < f_start and s_end == f_end:
        # case 4
        return []
    elif s_start < f_start and s_end > f_end:
        # case 5
        return []
    elif s_start == f_start and s_end == f_start:
        # case 6
        raise Exception('s_start should not equal s_end')
    elif s_start == f_start and s_end < f_end:
        # case 7
        return [(s_end, f_end)]
    elif s_start == f_start and s_end == f_end:
        # case 8
        return []
    elif s_start == f_start and s_end > f_end:
        # case 9
        return []
    elif s_start < f_end and s_end == s_start:
        # case 10
        raise Exception('s_start should not equal s_end')
    elif s_start < f_end and s_end < f_end:
        # case 11
        return [(f_start, s_start), (s_end, f_end)]
    elif s_start < f_end and s_end == f_end:
        # case 12
        return [(f_start, s_start)]
    elif s_start < f_end and s_end > f_end:
        # case 13
        return [(f_start, s_start)]
    elif s_start == f_end and s_end == f_end:
        # case 14
        raise Exception('s_start should not equal s_end')
    elif s_start == f_end and s_end > f_end:
        # case 15
        return [(f_start, f_end)]
    elif s_start > f_end and s_end > f_start:
        # case 16
        return [(f_start, f_end)]
    else:
        raise Exception()


def partition_into_chunks(
    data_ranges: list[typing.Any], range_format: absorb.RangeFormat
) -> list[typing.Any]:
    if range_format == 'date':
        import datetime

        dates = []
        for data_range in data_ranges:
            delta = datetime.timedelta(days=1)
            start_date, end_date = data_range
            current = start_date
            while current <= end_date:
                dates.append(current)
                current = current + delta
        return dates
    else:
        raise NotImplementedError()



from __future__ import annotations

import os
import typing
import absorb

from . import names

if typing.TYPE_CHECKING:
    import types
    import polars as pl


def get_source_module(source: str) -> types.ModuleType:
    import importlib

    return importlib.import_module(
        'absorb.datasets.' + names._camel_to_snake(source)
    )


def get_table_class(source: str, table_name: str) -> type[absorb.Table]:
    return getattr(get_source_module(source), names._snake_to_camel(table_name))  # type: ignore


def get_sources(*, snake: bool = True) -> list[str]:
    import absorb.datasets

    sources = [
        filename.rsplit('.py', maxsplit=1)[0]
        for filename in os.listdir(absorb.datasets.__path__[0])
        if not filename.startswith('__')
    ]

    if not snake:
        sources = [names._snake_to_camel(source) for source in sources]

    return sources


def get_source_tables(source: str) -> list[type[absorb.Table]]:
    module = get_source_module(source)
    if hasattr(module, 'get_tables'):
        return module.get_tables()  # type: ignore
    else:
        return [
            value
            for key, value in vars(module).items()
            if isinstance(value, type) and issubclass(value, absorb.Table)
        ]


def get_table_classes() -> list[type[absorb.Table]]:
    return [
        table_class
        for source in get_sources()
        for table_class in get_source_tables(source)
    ]


def resolve_table(
    reference: absorb.TableReference,
    *,
    parameters: dict[str, typing.Any] | None = None,
) -> absorb.Table:
    if isinstance(reference, absorb.Table):
        return reference

    elif isinstance(reference, dict):
        return absorb.Table.instantiate(reference)

    elif isinstance(reference, str):
        source, table = reference.split('.')
        camel_table = names._snake_to_camel(table)
        snake_table = names._camel_to_snake(table)
        if parameters is None:
            parameters = {}
        tracked_table: absorb.TrackedTable = {
            'source_name': source,
            'table_name': snake_table,
            'table_class': 'absorb.datasets.' + source + '.' + camel_table,
            'parameters': parameters,
        }
        return absorb.Table.instantiate(tracked_table)

    else:
        raise Exception()


def get_available_tables(
    *, exclude_parameters: bool = False
) -> list[absorb.TrackedTable]:
    tables = []
    for source in get_sources():
        for table in get_source_tables(source):
            if exclude_parameters and len(table.parameter_types) > 0:
                continue
            snake_table = absorb.ops.names._camel_to_snake(table.__name__)
            tracked_table: absorb.TrackedTable = {
                'source_name': source,
                'table_name': snake_table,
                'table_class': 'absorb.datasets.'
                + source
                + '.'
                + table.__name__,
                'parameters': {},
            }
            tables.append(tracked_table)
    return tables


def get_tables_df(
    include_available_range: bool = False,
) -> pl.DataFrame:
    """WIP

    features wanted
    - is tracked
    - collected range
    - available range
    """
    import datetime
    import polars as pl

    rows: list[dict[str, typing.Any]] = []
    for table in absorb.ops.get_available_tables():
        row: dict[str, typing.Any] = dict(table)
        table_class = absorb.ops.get_table_class(
            source=table['source_name'],
            table_name=table['table_name'],
        )
        row['parameter_names'] = list(table_class.parameter_types.keys())

        if include_available_range:
            if len(table_class.parameter_types) > 0:
                pass
            else:
                instance = absorb.ops.resolve_table(table)
                if table_class.range_format == 'date':
                    start, end = instance.get_available_range()
                    if start is not None:
                        start = start.replace(tzinfo=datetime.timezone.utc)
                    if end is not None:
                        end = end.replace(tzinfo=datetime.timezone.utc)
                    row['available_start_time'] = start
                    row['available_end_time'] = end
    tables = pl.DataFrame(rows, orient='row').drop('parameters')

    #     tables.join(
    #         pl.DataFrame(absorb.ops.get_tracked_tables())
    #         .drop("parameters")
    #         .with_columns(tracked=True),
    #         on=["source_name", "table_name", "table_class"],
    #         how="left",
    #     ).with_columns(pl.col.tracked.fill_null(False))

    return tables



from __future__ import annotations

import typing
import absorb
from . import table_base

if typing.TYPE_CHECKING:
    T = typing.TypeVar('T')


class TablePaths(table_base.TableBase):
    def get_dir_path(self, warn: bool = True) -> str:
        return absorb.ops.paths.get_table_dir(
            source=self.source, table=self.name(), warn=warn
        )

    def get_glob(self, warn: bool = True) -> str:
        return self.get_file_path(glob=True, warn=warn)

    def get_file_path(
        self,
        data_range: typing.Any | None = None,
        glob: bool = False,
        warn: bool = True,
    ) -> str:
        if self.write_range == 'overwrite_all':
            data_range = 'all'
            range_format = None
        else:
            range_format = self.range_format
        return absorb.ops.paths.get_table_filepath(
            data_range=data_range,
            range_format=range_format,
            filename_template=self.filename_template,
            table=self.name(),
            source=self.source,
            parameters=self.parameters,
            glob=glob,
            warn=warn,
        )

    def get_file_paths(
        self, data_ranges: typing.Any, warn: bool = True
    ) -> list[str]:
        return absorb.ops.paths.get_table_filepaths(
            data_ranges=data_ranges,
            range_format=self.range_format,
            filename_template=self.filename_template,
            table=self.name(),
            source=self.source,
            parameters=self.parameters,
            warn=warn,
        )

    def get_file_name(
        self, data_range: typing.Any, *, glob: bool = False
    ) -> str:
        if self.write_range == 'overwrite_all':
            data_range = 'all'
            range_format = None
        else:
            range_format = self.range_format
        return absorb.ops.paths.get_table_filename(
            data_range=data_range,
            range_format=range_format,
            filename_template=self.filename_template,
            table=self.name(),
            source=self.source,
            parameters=self.parameters,
            glob=glob,
        )

    def parse_file_path(self, path: str) -> dict[str, typing.Any]:
        if self.write_range == 'overwrite_all':
            range_format = None
        else:
            range_format = self.range_format
        return absorb.ops.paths.parse_file_path(
            path=path,
            filename_template=self.filename_template,
            range_format=range_format,
        )



from __future__ import annotations

import typing
import absorb

if typing.TYPE_CHECKING:
    T = typing.TypeVar('T')

    import polars as pl


class TableBase:
    source: str
    write_range: typing.Literal[
        'append_only', 'overwrite_all', 'overwrite_chunks'
    ]
    range_format: absorb.types.RangeFormat
    index_by: typing.Literal['time', 'block', 'id']
    cadence: typing.Literal['daily', 'weekly', 'monthly', 'yearly'] | None
    parameter_types: dict[str, typing.Any] = {}
    default_parameters: dict[str, typing.Any] = {}
    parameters: dict[str, typing.Any] = {}
    static_parameters: list[str] = []
    filename_template = '{source}__{table}__{data_range}.parquet'
    name_template: absorb.types.NameTemplate = {}

    def __init__(self, parameters: dict[str, typing.Any] | None = None):
        # set parameters
        if parameters is None:
            parameters = {}
        else:
            parameters = parameters.copy()
        for parameter in parameters.keys():
            if parameter in self.static_parameters:
                raise Exception('cannot change parameter: ' + parameter)
        parameters = dict(self.parameters, **parameters)
        for key, value in self.default_parameters.items():
            parameters.setdefault(key, value)
        if set(parameters.keys()) != set(self.parameter_types.keys()):
            raise Exception(
                self.name() + ': parameters must match parameter_types spec'
            )
        self.parameters = parameters

    def get_schema(self) -> dict[str, type[pl.DataType] | pl.DataType]:
        raise NotImplementedError()

    @classmethod
    def class_name(
        cls,
        allow_generic: bool = False,
        parameters: dict[str, typing.Any] | None = None,
    ) -> str:
        if parameters is not None:
            parameters = dict(
                cls.default_parameters, **cls.parameters, **parameters
            )
        else:
            parameters = dict(cls.default_parameters, **cls.parameters)
        return absorb.ops.get_table_name(
            base_name=absorb.ops.names._camel_to_snake(cls.__name__),
            name_template=cls.name_template,
            parameter_types=cls.parameter_types,
            parameters=parameters,
            default_parameters=cls.default_parameters,
            static_parameters=cls.static_parameters,
            allow_generic=allow_generic,
        )

    def name(self) -> str:
        return absorb.ops.get_table_name(
            base_name=absorb.ops.names._camel_to_snake(type(self).__name__),
            name_template=self.name_template,
            parameter_types=self.parameter_types,
            parameters=self.parameters,
            default_parameters=self.default_parameters,
            static_parameters=self.static_parameters,
        )

    # defaults

    def get_default_data_range(self) -> typing.Any:
        raise NotImplementedError()

    @staticmethod
    def instantiate(dataset: absorb.TrackedTable) -> absorb.Table:
        import importlib

        module_name, class_name = dataset['table_class'].rsplit('.', maxsplit=1)
        module = importlib.import_module(module_name)
        cls = getattr(module, class_name)
        return cls(parameters=dataset['parameters'])  # type: ignore



# https://docs.growthepie.xyz/api

from __future__ import annotations

import absorb

import typing

if typing.TYPE_CHECKING:
    import polars as pl


class Metrics(absorb.Table):
    source = 'growthepie'
    write_range = 'overwrite_all'
    range_format = 'date_range'

    def get_schema(self) -> dict[str, type[pl.DataType] | pl.DataType]:
        df = self.collect_chunk(None)
        if df is not None:
            return dict(df.schema)
        else:
            raise Exception('data not found')

    def collect_chunk(self, data_range: typing.Any) -> pl.DataFrame | None:
        import requests
        import polars as pl

        url = 'https://api.growthepie.xyz/v1/fundamentals_full.json'
        response = requests.get(url)
        data = response.json()
        return (
            pl.DataFrame(data)
            .with_columns(date=pl.col.date.str.to_date().cast(pl.Datetime))
            .rename({'origin_key': 'network'})
            .pivot(on='metric_key', index=['date', 'network'], values='value')
            .sort('date', 'network')
        )

    def get_available_range(self) -> typing.Any:
        import datetime
        import requests
        import polars as pl

        first = datetime.datetime(year=2021, month=6, day=1)

        url = 'https://api.growthepie.xyz/v1/fundamentals.json'
        response = requests.get(url)
        data = response.json()
        last_str: str = pl.DataFrame(data)['date'].max()  # type: ignore
        last = datetime.datetime.strptime(last_str, '%Y-%m-%d')

        return [first, last]
